#ifndef NAC_H
#define NAC_H

#include <map>
#include <vector>
#include <string>
#include <algorithm>

#include "linguisticgraph/linguisticgraph.h"

class NAC
{
public:
    NAC();
    virtual ~NAC(){}

    virtual bool keep(LinguisticGraph *sentence,
                               const std::map<unsigned, unsigned> &nodes,
                               const std::map<unsigned, unsigned> &edges) = 0;

    virtual ostream &output(ostream& stream) = 0;
    friend ostream &operator<<(ostream &cout,NAC &nac);
};


class FeatureNAC: public NAC
{
public:
    enum Mode{Node = 0, Edge};

    FeatureNAC(unsigned id, Mode mode);

    void setId(unsigned id){m_id = id;}
    void addFeature(std::string key, std::string value);

    bool keep(LinguisticGraph *sentence,
              const std::map<unsigned, unsigned> &nodes,
              const std::map<unsigned, unsigned> &edges);

    ostream &output(ostream& stream);

protected:
    unsigned m_id;
    Mode m_mode;

private:
    bool test(const LinguisticNodeProperties &node);
    bool test(const LinguisticEdgeProperties &edge);
private:
    std::multimap<std::string, std::string> m_features;
};

class GraphNAC: public NAC
{
    typedef std::pair< std::set<unsigned>, std::set<unsigned> > SetPair;
public:
    GraphNAC(LinguisticGraph *graph, SetPair ids );
    ~GraphNAC();

    bool keep(LinguisticGraph *sentence,
              const std::map<unsigned, unsigned> &nodes,
              const std::map<unsigned, unsigned> &edges);

    ostream &output(ostream& stream);

private:
    LinguisticGraph *m_graph;
    SetPair m_ids;
};

class PrecedenceNAC: public NAC
{
public:
    PrecedenceNAC(std::vector<unsigned> ids);
    ~PrecedenceNAC();

    bool keep(LinguisticGraph *sentence,
              const std::map<unsigned, unsigned> &nodes,
              const std::map<unsigned, unsigned> &edges);

    ostream &output(ostream &stream);

private:
    std::vector<unsigned> m_ids;
};

#endif //NAC_H
