#ifndef REWRITER_H
#define REWRITER_H

#include <map>
#include <string>
#include <vector>
#include <algorithm>

#include <linguisticgraph/linguisticgraph.h>
#include "ogrecomponents/rule.h"
#include "ogrecomponents/compiler.h"

namespace ogrecomponents{

using namespace std;

class Rewriter
{
    typedef vector<map<unsigned, unsigned> > MatchesType;
    typedef std::pair<MatchesType, MatchesType> MatchesPairType;
public:
    Rewriter(Verbosity v = NONE, bool onlymatch = false)
        :m_verbose(v), m_onlymatch(onlymatch)
    {}

    bool rewrite(LinguisticGraph *sentence,
                 LinguisticGraph *result,
                 Compiler *compiler);

protected:
    typedef std::vector< std::pair<unsigned, Constraint*> > ConstraintVector;

    MatchesPairType testNACs(Rule *rule,
                  LinguisticGraph *sentence,
                  vector<map<unsigned, unsigned> >&nodes,
                  vector<map<unsigned, unsigned> >&edges);

    MatchesPairType match(
            Rule *rule,
            LinguisticGraph *sentence
   );

    bool firstStageRewriting(LinguisticGraph *result,
                             Rule *rule,
                             const vector<map<unsigned, unsigned> > &matching_nodes,
                             const vector<map<unsigned, unsigned> > &matching_edges);

    void makeConstraint(LinguisticGraph *result, Rule *rule,
                        const vector<map<unsigned, unsigned> > &nodes,
                        const vector<map<unsigned, unsigned> > &edges,
                        ConstraintVector &constraints);

    bool secondStageRewriting(LinguisticGraph *result,
                              const std::vector<std::pair<unsigned, Constraint*> >& constraints);

private:
    void createMatchedSubgraph();

private:
    Verbosity m_verbose;
    bool m_onlymatch;
};

}

#endif //REWRITER_H
