#ifndef RULE_H
#define RULE_H

#include <linguisticgraph/linguisticgraph.h>
#include <vector>
#include <boost/variant.hpp>

#include "ogrecomponents/constraint.h"
#include "ogrecomponents/command.h"
#include "ogrecomponents/nac.h"

namespace ogrecomponents{

class Rule
{    
public:    
    typedef std::vector<Command*> CommandVector;
    typedef CommandVector::const_iterator CommandVectorConstIter;

    typedef boost::variant<unsigned, std::pair<unsigned, unsigned> > ConstraintIdType;
    typedef std::vector< std::pair<ConstraintIdType, Constraint*> > ConstraintVector;
    typedef ConstraintVector::const_iterator ConstraintVectorConstIter;

    typedef std::vector<NAC*> NacVector;
    typedef NacVector::const_iterator NacVectorConstIter;

    Rule():m_graph(NULL){}

    ~Rule();

    void setName(std::string name){m_name = name;}
    const std::string& name()const {return m_name;}

    void setGraph(LinguisticGraph *graph){m_graph = graph;}
    LinguisticGraph *graph()const {return m_graph;}

    bool add_node_var(std::string var, unsigned id);
    std::pair<unsigned,bool> get_node_id(std::string var) const;

    bool add_edge_var(std::string var, unsigned id);
    std::pair<unsigned,bool> get_edge_id(std::string var) const;

    void addCommand(Command *c){m_commands.push_back(c);}

    std::pair<CommandVectorConstIter, CommandVectorConstIter> commands()const
    {return make_pair(m_commands.begin(), m_commands.end());}

    unsigned commandsCount()const {return m_commands.size();}


    void addConstraint(ConstraintIdType eid, Constraint *c)
    {m_constraints.push_back(make_pair(eid, c));}

    std::pair<ConstraintVectorConstIter, ConstraintVectorConstIter> constraints()const    
    {return make_pair(m_constraints.begin(), m_constraints.end());}    

    unsigned edgesWithConstraintCount()const {return m_constraints.size();}

    void addNAC(NAC *nac){m_nacs.push_back(nac);}

    std::pair<NacVectorConstIter, NacVectorConstIter> NACs()const
    {return make_pair(m_nacs.begin(), m_nacs.end());}

    unsigned NACCount()const{return m_nacs.size();}

private:
    typedef std::map<std::string, unsigned> MapVars;
    typedef MapVars::const_iterator MapVarsConstIter;
    typedef MapVars::iterator MapVarsIter;

    std::string m_name;
    LinguisticGraph *m_graph;
    ConstraintVector m_constraints;
    CommandVector m_commands;
    NacVector m_nacs;
    MapVars m_node_vars, m_edge_vars;


};

}

#endif
