#include "ogrecomponents/command.h"
#include <linguisticgraph/matcher.h>
#include <boost/algorithm/string.hpp>
#include <boost/assign.hpp>

namespace ogrecomponents{

/*int Command::find_edge(unsigned eid_rule,
        const std::map<unsigned, unsigned>& matching,
        LinguisticGraph *rule, LinguisticGraph *sentence)
{
    unsigned src_in_rule = rule->nodeProperties(rule->sourceNode(eid_rule)).id;
    unsigned src_in_sentence = matching.find(src_in_rule)->second;

    LinguisticEdgeProperties pedge_in_rule = rule->edgeProperties(eid_rule);

    LinguisticGraph::out_edge_iter oit, eoit;
    for(boost::tie(oit, eoit) = sentence->outEdges(src_in_sentence); oit != eoit; ++oit)
    {
        LinguisticEdgeProperties possible_edge = sentence->edgeProperties(*oit);
        if(pedge_in_rule == possible_edge)
            return possible_edge.id;
    }
    return -1;
}*/

AddEdgeCommand::~AddEdgeCommand(){}

void AddEdgeCommand::add_feature(string key, string value)
{
    m_features.insert(make_pair(key, value));
}

bool AddEdgeCommand::process(const std::map<unsigned, unsigned> &matching_nodes,
                             const std::map<unsigned, unsigned> &matching_edges,
                             LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    unsigned src_in_sentence = matching_nodes.find(m_src)->second,
            tar_in_sentence = matching_nodes.find(m_tar)->second;

    std::string final = "", canon = "";
    std::string surfacic = "y", deep = "y";

    for(MapFeaturesConstIter it = m_features.begin(); it != m_features.end(); ++it)
    {
        if(it->first == "surfacic")
            surfacic = it->second;
        else if(it->first == "deep")
            deep     = it->second;
        else if(it->first == "label")
        {
            std::string label = it->second;
            if(linggraph::isCanonicalLabel(label))
                canon = label;
            else if(linggraph::isFinalLabel(label))
                final = label;
            else if(linggraph::isBothLabel(label))
            {
                std::vector<std::string> parts;
                boost::algorithm::split(parts, label, boost::algorithm::is_any_of(":"));
                if(parts.size() == 2)
                {
                    final = parts[0];
                    canon = parts[1];
                }
                else
                    final = label;
            }
            else
                final = label;
        }
    }

    LinguisticEdgeProperties edge(sentence->lastEdgeId()+1, final, canon);
    edge.setSurfacicState(surfacic);
    edge.setDeepState(deep);

    //Test to see if edge are identical
    LinguisticGraph::out_edge_iter oeit, oeend;
    for(boost::tie(oeit, oeend) = sentence->outEdges(src_in_sentence);
        oeit != oeend; ++oeit)
    {
        //If the target node is different continue to iterate, because the edge
        //we want to add, can't be identical to this one
        const LinguisticNodeProperties &target = sentence->nodeProperties(sentence->targetNode(*oeit));
        if(target.id != tar_in_sentence)
            continue;

        const LinguisticEdgeProperties e = sentence->edgeProperties(*oeit);

        //Edge are strictly identical, don't add a new one
        if(e.relFinal() == edge.relFinal()
                && e.relCanonical() == edge.relCanonical()
                && e.type() == edge.type())
        {
            if(m_verbose == DEBUG){
                std::cerr << "Edge " << edge << " already exists (" << e << ") . Don't add it again" << std::endl;
            }
            return true;
        }
    }

    sentence->addEdge(src_in_sentence, tar_in_sentence, edge);

    return true;
}

RemoveCommand::~RemoveCommand(){}

bool RemoveCommand::process(const std::map<unsigned, unsigned> &matching_nodes,
                            const std::map<unsigned, unsigned> &matching_edges,
                            LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    if(m_mode == RemoveCommand::Node)
    {
        unsigned node_in_sentence = matching_nodes.find(m_id)->second;
        try{
            sentence->removeNode(node_in_sentence);
        }catch(std::out_of_range &e){
            e.what();
        }
    }
    else
    {
        unsigned edge_in_sentence = matching_edges.find(m_id)->second;
        //std::cerr << "Removing edge: " << edge_in_sentence << std::endl;
        try{
            sentence->removeEdge(edge_in_sentence);
        }catch(std::out_of_range &e){
            e.what();
            //std::cerr << e.what() << std::endl;
        }
    }

    return true;
}

bool MoveEdgeCommand::process(const std::map<unsigned, unsigned> &matching_nodes,
                              const std::map<unsigned, unsigned> &matching_edges,
                              LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    unsigned eid_in_sent = matching_edges.find(m_eid)->second;
    unsigned src_in_sent = matching_nodes.find(m_newSrc)->second,
            tar_in_sent = matching_nodes.find(m_newTar)->second;

    if(m_type == Clone){
        std::pair<unsigned, bool> result = sentence->cloneEdge(src_in_sent, tar_in_sent, eid_in_sent);
        unsigned new_id = result.first;
        eid_in_sent = new_id;

        if( !result.second )
            return false;

        sentence->edgeProperties(new_id).setOriginal(false);
    }else if(m_type == Move){
        sentence->changeSourceEdge(eid_in_sent, src_in_sent);
        sentence->changeTargetEdge(eid_in_sent, tar_in_sent);                
    }

    modifyFeatures(sentence, eid_in_sent);

    //We modify the features, so we need to be sure, that we don't duplicate same edges
    LinguisticGraph::out_edge_iter oit, end;
    bool same = false;
    for(boost::tie(oit, end) = sentence->outEdges(src_in_sent); oit != end; ++oit)
    {
        const LinguisticEdgeProperties &out = sentence->edgeProperties(*oit);
        const LinguisticNodeProperties &target = sentence->nodeProperties( sentence->targetNode(*oit) );

        if( out.compare( sentence->edgeProperties(eid_in_sent) )
                && out.id != eid_in_sent  && target.id == tar_in_sent)
        {
            same = true;
            break;
        }
    }

    if(same){
        sentence->removeEdge(eid_in_sent);
    }

    return true;
}

void MoveEdgeCommand::add_feature(std::string key, std::string value)
{
    m_features.insert( make_pair( key, value) );
}

void MoveEdgeCommand::modifyFeatures(LinguisticGraph *graph, unsigned eid){
    LinguisticEdgeProperties &prop = graph->edgeProperties(eid);

    for(MapFeaturesConstIter it = m_features.begin(); it != m_features.end(); ++it){
        string key = it->first, value = it->second;
        if(key == "label"){
            string canon = prop.relCanonical(), final = prop.relFinal();

            if(linggraph::isCanonicalLabel(value))
                canon = value;
            else if(linggraph::isFinalLabel(value))
                final = value;
            else
            {
                std::vector<std::string> parts;
                boost::algorithm::split(parts, value, boost::algorithm::is_any_of(":"));
                if(parts.size() == 2)
                {
                    final = parts[0];
                    canon = parts[1];
                }
                else
                    final = value;
            }

            prop.setRelCanonical(canon);
            prop.setRelFinal(final);
        }else if(key == "deep"){
            prop.setDeepState(value);
        }else if(key == "surfacic"){
            prop.setSurfacicState(value);
        }
    }
}

bool FeatureCommand::Visitor::operator ()(const std::set<std::string> &values)
{
    std::copy( values.begin(), values.end(), std::inserter(m_values, m_values.begin()) );   
    return true;
}

bool FeatureCommand::Visitor::operator ()(const boost::tuples::tuple<unsigned, std::string,
                                          FeatureCommand::Element> &values)
{
    std::string key = values.get<1>();
    if( values.get<2>() == FeatureCommand::Node )
    {
        if( matching_nodes.find(values.get<0>()) != matching_nodes.end() )
        {
            unsigned node_id = matching_nodes.find(values.get<0>())->second;
            const LinguisticNodeProperties &nProp = sentence->nodeProperties(node_id);
            m_values.insert( nProp.feature(key) );
        }
        else
        {
            std::cerr << "Unable to find node id " << values.get<0>()
                      << " for feature dictionary" << std::endl;
            return false;
        }
    }
    else
    {
        if( matching_edges.find(values.get<0>()) != matching_edges.end() )
        {
            unsigned edge_id = matching_edges.find(values.get<0>())->second;
            const LinguisticEdgeProperties &eProp = sentence->edgeProperties(edge_id);
            std::string value = "";
            if(key == "label")
            {
                if(eProp.relFinal().empty())
                {
                    if(eProp.relCanonical().empty())
                        value = "";
                    else
                        value = ":"+eProp.relCanonical();
                }
                else
                {
                    if(eProp.relCanonical().empty())
                        value = eProp.relFinal();
                    else
                        value = eProp.relFinal()+":"+eProp.relCanonical();
                }
            }
            else if(key == "status")
            {
                value = eProp.typeString();
            }          

            m_values.insert( value );
        }
        else
        {
            std::cerr << "Unable to find edge id "
                      << values.get<0>() << " for feature dictionary" << std::endl;
            return false;
        }
    }

    return true;
}

bool FeatureCommand::process(const std::map<unsigned, unsigned> &matching_nodes,
                             const std::map<unsigned, unsigned> &matching_edges,
                             LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    std::set<std::string> values;

    Visitor visitor(matching_nodes, matching_edges, sentence);
    bool success = boost::apply_visitor(visitor, m_values);
    if( !success )
        return false;
    values = visitor.values();

    if(m_element == Edge)
    {        
        try{
            LinguisticEdgeProperties& edge = sentence->edgeProperties(matching_edges.find(m_id)->second);
            feature_command_on_edge(edge, values);
        }catch(std::out_of_range&){
            return false;
        }
    }
    else
    {
        try{
            LinguisticNodeProperties& node = sentence->nodeProperties(matching_nodes.find(m_id)->second);
            feature_command_on_node(node, values);
        }catch(std::out_of_range&){
            return false;
        }
    }

    return true;
}

void FeatureCommand::feature_command_on_edge(LinguisticEdgeProperties &edge,
                                             const std::set<std::string> &values)
{   
    for(std::set<std::string>::const_iterator it = values.begin(); it != values.end(); ++it)
    {
        string val = *it, final = "", canon = "";
        if(m_key == "label")
        {
            if(linggraph::isCanonicalLabel(val))
                canon = val;
            else if(linggraph::isFinalLabel(val))
                final = val;
            else
            {
                std::vector<std::string> parts;
                boost::algorithm::split(parts, val, boost::algorithm::is_any_of(":"));
                if(parts.size() == 2)
                {
                    final = parts[0];
                    canon = parts[1];
                }
                else
                    final = val;
            }
        }       
        else if(m_key == "surfacic")
        {
            if( val == "y" )
                edge.setSurfacicState(true);
            else
                edge.setSurfacicState(false);
        }
        else if(m_key == "deep")
        {
            if( val == "y")
                edge.setDeepState(true);
            else
                edge.setDeepState(false);
        }

        switch(m_mode)
        {
        case Append:
            if(!final.empty())
                edge.addRelFinal(final);
            if(!canon.empty())
                edge.addRelCanonical(canon);
            break;

        case Add:
            if(!final.empty())
                edge.setRelFinal(final);
            if(!canon.empty())
                edge.setRelCanonical(canon);
            break;

        case Remove:
            std::cerr << "Remove feature is not implemented for edge" << std::endl;
            break;

        default:
            break; //Should never happen.
        }
    }
}

void FeatureCommand::feature_command_on_node(LinguisticNodeProperties &node,
                                             const std::set<std::string> &values)
{
    if(m_mode == FeatureCommand::Remove){

        if(m_key == "token")
            node.removeToken();
        else if(m_key == "lemma")
            node.removeLemma();
        else if(m_key == "hypercat")
            node.removeHyperpos();
        else if(m_key == "cat")
            node.removePos();
        else
            node.clearFeature(m_key);

    }else{
        for(std::set<std::string>::const_iterator it = values.begin();
            it != values.end(); ++it)
        {
            switch(m_mode){
            case Append:
                if(m_key == "token")
                    node.addToken(*it);
                else if(m_key == "lemma")
                    node.addLemma(*it);
                else if(m_key == "hypercat")
                    node.addHyperpos(*it);
                else if(m_key == "cat")
                    node.addPos(*it);
                else
                    node.addFeature(m_key, *it);
                break;
            case Add:
                if(m_key == "token")
                    node.setToken(*it);
                else if(m_key == "lemma")
                    node.setLemma(*it);
                else if(m_key == "hypercat")
                    node.setHyperpos(*it);
                else if(m_key == "cat")
                    node.setPos(*it);
                else
                    node.setFeature(m_key, *it);
                break;
            default:
                break;
            }
        }
    }
}

bool ModifyFeatureCommand::process(const std::map<unsigned, unsigned>& matching_nodes,
                                   const std::map<unsigned, unsigned>& matching_edges,
                                   LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    if(m_element == Node){
        try{
            unsigned id = matching_nodes.find(m_id)->second;
            LinguisticNodeProperties &node = sentence->nodeProperties(id);
            modify_on_node(node);
        }catch(std::out_of_range&){
            return false;
        }
    }else{
        try{
            unsigned id = matching_edges.find(m_id)->second;
            LinguisticEdgeProperties &edge = sentence->edgeProperties(id);
            modify_on_edge(edge);
        }catch(std::out_of_range&){
            return false;
        }
    }
    return true;
}

void ModifyFeatureCommand::modify_on_node(LinguisticNodeProperties &node)
{   
    if(m_key == "token" && m_condition == node.token()){
        node.setToken(m_replace);
    }else if(m_key == "lemma" && m_condition == node.lemma()){
        node.setLemma(m_replace);
    }else if(m_key == "hypercat" && m_condition == node.hyperpos()){
        node.setHyperpos(m_replace);
    }else if(m_key == "cat" && m_condition == node.pos()){
        node.setPos(m_replace);
    }else{
        if(m_condition == node.feature(m_key)){
            node.setFeature(m_key, m_replace);
        }
    }
}

void ModifyFeatureCommand::modify_on_edge(LinguisticEdgeProperties &edge){
    if(m_key == "label")
    {
        if(linggraph::isCanonicalLabel(m_replace)){
            boost::algorithm::replace_first(m_condition, ":", "");
            if(m_condition == edge.relCanonical())
                edge.setRelCanonical(m_replace);
        }
        else if(linggraph::isFinalLabel(m_replace)){
            if(m_condition == edge.relFinal()){
                edge.setRelFinal(m_replace);
            }
        }
        else
        {
            std::vector<std::string> parts;
            std::vector<std::string> cond_parts;
            boost::algorithm::split(parts, m_replace, boost::algorithm::is_any_of(":"));
            boost::algorithm::split(cond_parts, m_condition, boost::algorithm::is_any_of(":"));
            if(parts.size() == 2 && cond_parts.size() == 2)
            {
                if(cond_parts[0] == edge.relFinal() && cond_parts[1] == edge.relCanonical()){
                    edge.setRelFinal(parts[0]);
                    edge.setRelCanonical(parts[1]);
                }
            }
            else
                edge.setRelFinal(m_replace);
        }
    }
    else if(m_key == "status" && edge.type() == linggraph::string2type(m_condition)){
        edge.setDependencyType( linggraph::string2type(m_replace) );
    }
}


bool StatusEdgeCommand::process(const std::map<unsigned, unsigned>&,
                                const std::map<unsigned, unsigned>& matching_edges,
                                LinguisticGraph *sentence)
{
    if(m_verbose == DEBUG)
        std::cerr << m_name << " command processing" << std::endl;

    unsigned sent_id = matching_edges.find(m_id)->second;
    LinguisticEdgeProperties &edge = sentence->edgeProperties(sent_id);


    edge.setSurfacicState(m_surfacic);
    edge.setDeepState(m_deep);

    return true;
}

}

