#include <boost/foreach.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/erase.hpp>
#include <boost/algorithm/string/replace.hpp>
#include <boost/assign/list_of.hpp>

#include <algorithm>

#include "ogrecomponents/constraint.h"

namespace ogrecomponents{

Constraint::Constraint(){}

SimpleConstraint::SimpleConstraint(set<string> labels, std::map<std::string, std::string> features,
                                   Verbosity verbose)
    :m_labels(labels), m_features(features), m_verbose(verbose)
{}

SimpleConstraint::~SimpleConstraint()
{
    m_labels.clear();
}

set<pair<unsigned, std::string> > SimpleConstraint::solutions(LinguisticGraph *g, set<unsigned> edges)
{
    set< pair<unsigned, string> > _solutions;
    for(set<unsigned>::const_iterator it = edges.begin(); it != edges.end(); ++it)
    {
        string final = g->edgeProperties(*it).relFinal();
        string canonical = g->edgeProperties(*it).relCanonical();

        //set<string> inter;
        std::vector<std::string> labels = boost::assign::list_of(final)(":"+canonical)(final+":"+canonical);
        for(int i = 0; i < labels.size(); i++)
        {
            if( m_labels.find( labels[i] ) != m_labels.end() ){
                if(i == 0)
                    _solutions.insert( make_pair(*it, "f"));
                else if(i == 1)
                    _solutions.insert( make_pair(*it, "c"));
                else
                    _solutions.insert( make_pair(*it, "fc"));
            }

        }
        /*set<string> sentenceLabels = boost::assign::list_of(final);
                //(":"+canonical)(final+":"+canonical);

        std::set_intersection(sentenceLabels.begin(), sentenceLabels.end(),
                              m_labels.begin(), m_labels.end(),
                              std::inserter(inter, inter.end()));*/
        //if(inter.size() >= 1)
            //_solutions.insert(*it);
    }
    return _solutions;
}


/*RedirectUp::RedirectUp(set<string> labels,
                       Verbosity verbose)
    :SimpleConstraint(labels, verbose)
{}

bool RedirectUp::trigger(GraphWrapper *&g, string eid)
{
    if(m_verbose)
        std::cerr << "In redirect up for edge: " << eid << std::endl;
    GraphWrapper::NodesIterator src = g->sourceNode(eid), tar = g->targetNode(eid);
    unordered_set<string> out = g->getOutEdges(tar);

    //bool ok;
    unordered_set<string> solutions = getSolutions(g, out);
    //if(ok)
    if(solutions.size() > 0)
    {
        //Applying solutions
        BOOST_FOREACH(string solution, solutions){
            g->moveEdge(solution, src, g->targetNode(solution));
            set<string> values;
            values.insert("false");
            g->setEdgeProperty(solution, "origin", values);

            //Hack because I'm too tired
            changeStatus(g, solution, "surfacic", false);
            changeStatus(g, solution, "deep", true);
        }
        return true;
    }
    else
    {
        return false;
    }
}*/

RedirectDown::RedirectDown(set<string> labels, std::map<std::string, std::string> features,
                           Verbosity verbose)
    :SimpleConstraint(labels, features, verbose)
{}

bool RedirectDown::trigger(LinguisticGraph *g, unsigned eid)
{
    bool applied = false;
    if(m_verbose >= DEBUG){
        std::cerr << "In redirect down";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }

    LinguisticGraph::Node src = g->sourceNode(eid), tar = g->targetNode(eid);
    LinguisticGraph::in_edge_iter bein, eein;
    set<unsigned> inIds;

    for(boost::tie(bein, eein) = g->inEdges( g->nodeProperties(src).id );
        bein != eein; ++bein)
        inIds.insert(g->edgeProperties(*bein).id);

    set< pair<unsigned, string> > _solutions = solutions(g, inIds);
    if(_solutions.size() > 0)
    {
        //Applying solutions
        typedef pair<unsigned, string> IdLabel;
        BOOST_FOREACH(IdLabel solution, _solutions){
            unsigned newSource = g->nodeProperties(g->sourceNode(solution.first)).id;
            unsigned newTarget = g->nodeProperties(tar).id;

            if(hasEdge(newSource, newTarget, solution.first, g, true))
                continue;

            g->changeTargetEdge(solution.first, newTarget);
            fix_label( g->edgeProperties(solution.first), solution.second );

            /*if(!m_status.empty())
                g->edgeProperties(solution.first).setDependencyType(linggraph::string2type(m_status));*/

            insert_features(m_features, g->edgeProperties(solution.first));

            applied = true;
        }
    }   
    return applied;
}

ShareUp::ShareUp(set<string> labels, std::map<std::string, std::string> features,
                 Verbosity verbose)
    :SimpleConstraint(labels, features, verbose)
{}

bool ShareUp::trigger(LinguisticGraph *g, unsigned eid)
{
    /**
     * new_src ----> src ---> tar
     * What we want is a new edge between new_src and tar
     * We start by finding all possible edges from new_src to src
     * We clone the edge between new_src and tar
     */

    bool applied = false;

    if(m_verbose >= DEBUG){
        std::cerr << "In share up";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }

    LinguisticGraph::Node src = g->sourceNode(eid), tar = g->targetNode(eid);
    LinguisticGraph::out_edge_iter be_out, ee_out;
    set<unsigned> outIds;

    for(boost::tie(be_out, ee_out) = g->outEdges( g->nodeProperties(tar).id );
        be_out != ee_out; ++be_out)
        outIds.insert(g->edgeProperties(*be_out).id);

    set<pair<unsigned, string> > _solutions = solutions(g, outIds);
    if(_solutions.size() > 0)
    {
        typedef pair<unsigned, string> IdLabel;
        BOOST_FOREACH(IdLabel solution, _solutions){
            unsigned new_src = g->nodeProperties(src).id;
            unsigned new_tar = g->nodeProperties( g->targetNode(solution.first) ).id;

            if(hasEdge(new_src, new_tar, solution.first, g, true))
                continue;

            std::pair<unsigned,bool> result = g->cloneEdge(new_src, new_tar, solution.first);
            unsigned new_id = result.first;
            if(!result.second){
                continue;
            }

            g->edgeProperties(new_id).setOriginal(false);
            fix_label( g->edgeProperties(new_id), solution.second );

            insert_features(m_features, g->edgeProperties(new_id));
            /*if(!m_status.empty())
                g->edgeProperties(new_id).setDependencyType(linggraph::string2type(m_status));*/

            m_historic.insert(new_id);
            applied = true;
        }
    }
    return applied;
}

ShareDown::ShareDown(set<string> labels, std::map<std::string, std::string> features,
                     Verbosity verbose)
    :SimpleConstraint(labels, features, verbose)
{}

bool ShareDown::trigger(LinguisticGraph *g, unsigned eid)
{
    bool applied = false;

    if(m_verbose >= DEBUG){
        std::cerr << "In share down";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }

    LinguisticGraph::Node src = g->sourceNode(eid),
            tar = g->targetNode(eid);

    LinguisticGraph::in_edge_iter bein, eein;
    set<unsigned> inIds;

    for(boost::tie(bein, eein) = g->inEdges( g->nodeProperties(src).id );
        bein != eein; ++bein)
        inIds.insert(g->edgeProperties(*bein).id);

    set< pair<unsigned, string> > _solutions = solutions(g, inIds);

    if(_solutions.size() > 0)
    {
        typedef pair<unsigned, string> IdLabel;
        BOOST_FOREACH(IdLabel solution, _solutions){
            unsigned new_src = g->nodeProperties(g->sourceNode(solution.first)).id;
            unsigned new_tar = g->nodeProperties(tar).id;

            if(hasEdge(new_src, new_tar, solution.first, g, true))
                continue;

            std::pair<unsigned, bool> result = g->cloneEdge(new_src, new_tar, solution.first);
            unsigned new_id = result.first;

            if( !result.second ){
                continue;
            }

            g->edgeProperties(new_id).setOriginal(false);
            fix_label( g->edgeProperties(new_id), solution.second );

            insert_features(m_features, g->edgeProperties(new_id));
            //g->edgeProperties(new_id).setDependencyType(linggraph::string2type(newone));

            m_historic.insert(new_id);
            applied = true;
        }
        return applied;
    }
    return applied; //false
}

AddFeature::AddFeature(set<string> labels, string final, string canonical,
                       std::map<std::string, std::string> features, Verbosity verbose)
    :SimpleConstraint(labels, features, verbose), m_final(final), m_canonical(canonical)
{}

bool AddFeature::trigger(LinguisticGraph *g, unsigned eid)
{
    bool applied = false;
    if(m_verbose >= DEBUG){
        std::cerr << "In add feature";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }

    LinguisticGraph::Node src = g->sourceNode(eid),
            tar = g->targetNode(eid);

    LinguisticGraph::out_edge_iter beout, eeout;
    set<unsigned> outIds;

    for(boost::tie(beout, eeout) = g->outEdges( g->nodeProperties(src).id );
        beout != eeout; ++beout)
        outIds.insert(g->edgeProperties(*beout).id);

    set< pair<unsigned, string> > _solutions = solutions(g, outIds);

    if(_solutions.size() > 0)
    {
        //Find all out edges with final label m_final and add the label m_canonical
        for(boost::tie(beout, eeout) = g->outEdges( g->nodeProperties(src).id );
            beout != eeout; ++beout)
        {
            if(m_historic.find(g->edgeProperties(*beout).id) != m_historic.end()){
                continue;                       
            }

            if(g->edgeProperties(*beout).relFinal() == m_final)
            {
                g->edgeProperties(*beout).setRelCanonical(m_canonical);

                insert_features(m_features, g->edgeProperties(*beout));
                //g->edgeProperties(*beout).setDependencyType(linggraph::string2type(m_status));

                m_historic.insert(g->edgeProperties(*beout).id);
            }
        }        
    }
    return applied;
}

AutomatonConstraint::AutomatonConstraint(DFA *dfa, string status, Verbosity verbose)
    :m_dfa(dfa), m_status(status), m_verbose(verbose)
{}

bool AutomatonConstraint::inspectDFA(LinguisticGraph *graph, set<unsigned> & ids,
                                map<unsigned, string> &solutions, Direction dir)
{
    while( ids.size() != 0 && !m_dfa->isFinal(m_dfa->current()) )
    {
        solutions.clear();
        for(set<unsigned>::const_iterator it = ids.begin(); it != ids.end(); ++it)
        {
            LinguisticEdgeProperties prop = graph->edgeProperties(*it);
            std::string finalLabel = prop.relFinal();
            std::string canoLabel  = ":"+prop.relCanonical();
            std::string finalAndCanoLabel = finalLabel + ":" + canoLabel;

            vector<std::string> possibilities = boost::assign::list_of(finalLabel)(canoLabel)(finalAndCanoLabel);
            for(int i = 0; i < possibilities.size(); i++)
            {
                std::string label = possibilities[i];
                try{
                    //Test if it exists a transition between the current automaton state
                    // and another state with the input 'label'
                    m_dfa->testTransition(label);
                    solutions.insert(make_pair(*it, label));
                }catch(DFAError *)
                {}
            }
        }

        if(solutions.size() != 1)
            return false;

        map<unsigned, std::string>::iterator it = solutions.begin();
        std::string label = it->second;
        m_dfa->next(label); //Automaton head goes forward

        if(m_verbose >= DEBUG)
            std::cerr << "An edge with label: " << label << " was found, so automaton goes forward" << std::endl;

        //We expect we can continue to move down
        ids.clear();

        if(dir == DOWN)
        {
            const LinguisticNodeProperties& nodeProp = graph->nodeProperties(graph->sourceNode(it->first));
            LinguisticGraph::in_edge_iter bein, eein;
            for(boost::tie(bein, eein) = graph->inEdges( nodeProp.id ); bein != eein; ++bein)
                ids.insert( graph->edgeProperties(*bein).id );
        }
        else
        {
            std::cerr << "UP direction is not implemented yet. Abort !!" << std::endl;
        }
    }

    return true;
}

/*MoveUp::MoveUp(DFA *dfa, Verbosity verbose)
    :m_dfa(dfa), m_verbose(verbose)
{}

bool MoveUp::trigger(GraphWrapper *&g, string eid)
{
    if(m_verbose)
        std::cerr << "In move_up for edge: " << eid << std::endl;
    GraphWrapper::NodesIterator src = g->sourceNode(eid), tar = g->targetNode(eid);

    unordered_set<string> in = g->getInEdges(src);
    unordered_set<string> solutions;
    while(!m_dfa->isFinal(m_dfa->current()) && in.size() != 0)
    {
        solutions.clear();
        for(unordered_set<string>::const_iterator it = in.begin(); it != in.end(); ++it)
        {
            string label = g->edgeProperty(*it, "label");
            try{
                //Test if it exists a transition between the current automaton state
                // and another state with the input 'label'
                m_dfa->testTransition(label);
                solutions.insert(*it);
            }catch(DFAError *e)
            {}
        }

        if(solutions.size() != 1)
            return false;

        string label = g->edgeProperty(*solutions.begin(), "label");
        m_dfa->next(label); //Automaton head goes forward

        if(m_verbose)
            std::cerr << "An edge with label: " << label << " was found, so automaton goes forward" << std::endl;

        //We expect we can continue to move down
        in = g->getInEdges( g->sourceNode(*solutions.begin()) );
    }

    if(solutions.size() != 1)
        return false;

    if(m_dfa->isFinal(m_dfa->current()))
    {
        if(g->sourceNode(*solutions.begin())->first != tar->first)
        {
            if(m_verbose)
                std::cerr << "We have a final configuration for move_down, so we move the edge" << std::endl;
            g->moveEdge(eid, g->sourceNode(*solutions.begin()), tar);

            set<string> values;
            values.insert("false");
            g->setEdgeProperty(eid, "origin", values);
            return true;
        }
        return false;
    }
    else
    {
        //Impossible to find a final state with the path in the graph
        return false;
    }
}*/

MoveDown::~MoveDown()
{
    delete m_dfa;
}

MoveDown::MoveDown(DFA *dfa, string status, Verbosity verbose)
    :AutomatonConstraint(dfa, status, verbose)
{}

bool MoveDown::trigger(LinguisticGraph *g, unsigned eid)
{
    if(m_verbose >= DEBUG){
        std::cerr << "In move down";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }


    LinguisticGraph::Node src = g->sourceNode(eid); //tar = g->targetNode(eid);

    LinguisticGraph::in_edge_iter bein, eein;
    set<unsigned> inIds;

    for(boost::tie(bein, eein) = g->inEdges( g->nodeProperties(src).id );
        bein != eein; ++bein)
        inIds.insert(g->edgeProperties(*bein).id);

    map<unsigned, string> solutions;
    bool success = this->inspectDFA(g, inIds, solutions, DOWN);
    if( !success )
        return success;

    if(solutions.size() != 1)
        return false;

    if(m_dfa->isFinal(m_dfa->current()))
    {
        map<unsigned, string>::iterator it = solutions.begin();
        unsigned new_src = g->nodeProperties(src).id;
        unsigned new_tar = g->nodeProperties( g->sourceNode(it->first) ).id;

        if(hasEdge(new_src, new_tar, it->first, g, true))
            return false;

        if(m_verbose >= DEBUG)
            std::cerr << "We have a final configuration for move_down, so we move the edge" << std::endl;

        g->changeTargetEdge(eid, new_tar);

        std::cerr << "change is ok" << std::endl;

        if(!m_status.empty())
            g->edgeProperties(eid).setDependencyType(linggraph::string2type(m_status));

        std::cerr << "dependency is ok" << std::endl;

        m_dfa->reset();
        return false;
    }
    else
    {
        //Unable to find a final state with the path in the graph
        m_dfa->reset();
        return false;
    }
}


/**
 * LongShareDown Constraint
 *
 */

MoveShareDown::MoveShareDown(DFA *dfa, string status, Verbosity verbose)
    :AutomatonConstraint(dfa, status, verbose)
{}

MoveShareDown::~MoveShareDown()
{
    delete m_dfa;
}

bool MoveShareDown::trigger(LinguisticGraph *g, unsigned eid)
{
    if(m_verbose >= DEBUG){
        std::cerr << "In move share down";
        std::cerr << " - edge: " << g->edgeProperties(eid) << std::endl;
    }

    LinguisticGraph::Node src = g->sourceNode(eid); //tar = g->targetNode(eid);

    LinguisticGraph::in_edge_iter bein, eein;
    set<unsigned> inIds;

    for(boost::tie(bein, eein) = g->inEdges( g->nodeProperties(src).id );
        bein != eein; ++bein)
        inIds.insert(g->edgeProperties(*bein).id);

    map<unsigned, string> solutions;
    bool success = this->inspectDFA(g, inIds, solutions, DOWN);
    if( !success )
        return success;

    if(solutions.size() != 1)
        return false;

    if(m_dfa->isFinal(m_dfa->current()))
    {
        map<unsigned, string>::iterator it = solutions.begin();
        unsigned new_src = g->nodeProperties(src).id;
        unsigned new_tar = g->nodeProperties( g->sourceNode(it->first) ).id;

        if(hasEdge(new_src, new_tar, it->first, g, true)){
            std::cerr << "edge: " << g->edgeProperties(it->first) << " already exists" << std::endl;
            return false;
        }

        if(m_verbose >= DEBUG)
            std::cerr << "We have a final configuration for move_down, so we move the edge" << std::endl;

        std::pair<unsigned, bool> result = g->cloneEdge( new_src, new_tar, eid );
        unsigned new_id = result.first;

        if( !result.second )
            return false;

        g->edgeProperties(new_id).setOriginal(false);

        if(!m_status.empty()){
            std::string original = "", newone  = "";
            evaluateStatus(m_status, original, newone);
            if(!newone.empty())
                g->edgeProperties(new_id).setDependencyType(linggraph::string2type(newone));
            if(!original.empty())
                g->edgeProperties(eid).setDependencyType(linggraph::string2type(original));
        }

        m_historic.insert(new_id);
        m_dfa->reset();
        return false;
    }
    else
    {
        //Unable to find a final state with the path in the graph
        m_dfa->reset();
        return false;
    }
}


}
