#include "ogrecomponents/dfa.h"

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <algorithm>

namespace ogrecomponents{

DFA::DFA(unordered_set<string> alpha)
{
    startState = -1;
    alphabet = alpha;
    //createAlphabet();
}

bool DFA::isNot(string input)
{
    if(algorithm::starts_with(input, "__not__"))
        return true;
    return false;
}

bool DFA::isDisjunction(string input)
{
    if(algorithm::contains(input, "|"))
        return true;
    return false;
}

vector<string> DFA::disjunct(string input)
{
    vector<string> splitted;
    return boost::algorithm::split(splitted, input, boost::algorithm::is_any_of("|"));
}

string DFA::removeLogical(string input)
{
    algorithm::erase_first(input, "__not__");
    return input;
}

void DFA::addTransition(unsigned src, string input, unsigned dest)
{
    if(startState == -1)
    {
        startState = (int)src;
        currentState = src;
    }

    TransitionsIterator it = transitions.find(src);
    unordered_map<string, unsigned> end;
    if(it != transitions.end())
        end = it->second;

    if(isNot(input))
    {
        /*input = removeLogical(input);
        unordered_set<string> _not; _not.insert(input);
        unordered_set<string> allCorrespondingLabels;

        std::set_difference(alphabet.begin(), alphabet.end(),
                            _not.begin(), _not.end(),
                            std::inserter(allCorrespondingLabels,
                                          allCorrespondingLabels.end()));

        //uset_substract<string>(allCorrespondingLabels, alphabet, _not);*/

        end.insert(make_pair(input, dest));

        /*for(unordered_set<string>::iterator it = allCorrespondingLabels.begin();
            it != allCorrespondingLabels.end(); ++it)
        {
            end.insert(make_pair(*it, dest));
        }*/

    }else{

        vector<string> labels;
        if(isDisjunction(input))
            labels = disjunct(input);
        else
            labels.push_back(input);

        for(vector<string>::const_iterator it = labels.begin(); it != labels.end(); ++it)
            end.insert(make_pair(*it, dest));
    }

    transitions[src] = end;
}

bool DFA::isFinal(unsigned state) const
{
    if(finalStates.find(state) != finalStates.end())
        return true;
    else
        return false;
}

void DFA::addFinalState(unsigned state)
{
    finalStates.insert(state);
}

unsigned DFA::nextState(unsigned src, string input)
{
    typedef unordered_map<string, unsigned> StrToUMap;

    TransitionsIterator it = transitions.find(src);
    if(it == transitions.end())
    {
        format fmter("State %1% does not exist");
        fmter % src;
        throw new DFAError(fmter.str());
    }
    else
    {        
        StrToUMap end = it->second;
        for(StrToUMap::const_iterator strit = end.begin(); strit != end.end(); ++strit){
            string label = strit->first;
            //Test if label correspond to a __not__ label
            if(isNot(label)){
                //std::cerr << "isNot DFA ok" << std::endl;
                label = removeLogical(label);
                //std::cerr << "Label to find: " << label << std::endl;
                if(label != input)
                {
                    /*std::cerr << input << " is not " << label << std::endl;
                    std::cerr << "next state will be: " << strit->second << std::endl;*/
                    currentState = strit->second;
                    return currentState;
                }
            }else{
                if(label == input){
                    currentState = strit->second;
                    return currentState;
                }
            }
        }

        format fmter("There is no transition from %1% with input %2%");
        fmter % src % input;
        throw new DFAError(fmter.str());
    }
}

unsigned DFA::testTransition(string input)
{
    unsigned c = currentState;
    unsigned n = nextState(currentState, input);
    currentState = c;
    return n;
}

unsigned DFA::next(string input)
{
    return nextState(currentState, input);
}

}
