#include "ogrecomponents/rule.h"

namespace ogrecomponents{

Rule::~Rule()
{
    if(m_graph)
        delete m_graph;

    for(CommandVector::iterator it = m_commands.begin(); it != m_commands.end(); ++it)
    {
        Command *c = *it;
        delete c;
    }

    for(ConstraintVector::iterator it = m_constraints.begin(); it != m_constraints.end(); ++it)
    {
        std::pair<ConstraintIdType, Constraint*> p = *it;
        Constraint *c = p.second;
        delete c;
    }
}

bool Rule::add_node_var(string var, unsigned id)
{
    if(m_node_vars.find(var) != m_node_vars.end())
        return false;
    m_node_vars.insert(make_pair(var, id));
    return true;
}


bool Rule::add_edge_var(string var, unsigned id)
{
    if(m_edge_vars.find(var) != m_edge_vars.end())
        return false;
    m_edge_vars.insert(make_pair(var, id));
    return true;
}

std::pair<unsigned, bool> Rule::get_edge_id(string var) const
{
    MapVarsConstIter found = m_edge_vars.find(var);
    if(found != m_edge_vars.end())
        return make_pair(found->second, true);
    else
        return make_pair(0, false);
}

std::pair<unsigned, bool> Rule::get_node_id(string var) const
{
    MapVarsConstIter found = m_node_vars.find(var);
    if(found != m_node_vars.end())
        return make_pair(found->second, true);
    else
       return make_pair(0, false);
}

}
