#include <boost/test/unit_test.hpp>
#include <boost/assign/list_of.hpp>
#include <iostream>

#include <linguisticgraph/linguisticgraph.h>

#include "ogrecomponents/constraint.h"

using namespace ogrecomponents;

class ConstraintFixture
{
public:
    ConstraintFixture()
    {
        makeShareDownGraph();
        makeShareUpGraph();
        makeRedirectDownGraph();
        makeMoveDownGraph();
        makeAddFeatureGraph();
    }

    void makeShareDownGraph()
    {
        shareDownG = new LinguisticGraph();

        LinguisticNodeProperties jean(0, "Jean", "Jean", "N", "NPP"),
                aime(1, "aime", "aimer", "V", "V"),
                manger(2, "manger", "manger", "V", "VINF"),
                des(3, "des", "le", "D", "DET"),
                pommes(4, "pommes", "pomme", "N", "NC");

        LinguisticEdgeProperties suj(0, "suj", "suj"),
                obj(1, "obj", "obj"),
                det(2, "det", "det"),
                obj2(3, "obj", "obj");

        shareDownG->addNode(jean.id, jean);
        shareDownG->addNode(aime.id, aime);
        shareDownG->addNode(manger.id, manger);
        shareDownG->addNode(des.id, des);
        shareDownG->addNode(pommes.id, pommes);

        // aime -suj-> Jean
        shareDownG->addEdge(aime.id, jean.id, suj);
        // aime -obj-> manger
        shareDownG->addEdge(aime.id, manger.id, obj);
        // manger -obj-> pommes
        shareDownG->addEdge(manger.id, pommes.id, obj2);
        // pommes -det-> des
        shareDownG->addEdge(pommes.id, det.id, det);
    }

    void makeShareUpGraph()
    {
        shareUpG = new LinguisticGraph;

        LinguisticNodeProperties jean(0, "Jean", "Jean", "N", "NPP"),
                aime(1, "aime", "aimer", "V", "V"),
                manger(2, "manger", "manger", "V", "VINF"),
                des(3, "des", "le", "D", "DET"),
                pommes(4, "pommes", "pomme", "N", "NC");

        LinguisticEdgeProperties suj(0, "suj", "suj"),
                obj(1, "obj", "obj"),
                det(2, "det", "det"),
                obj2(3, "obj", "obj"),
                constraint(4, "__ogre_constraints__", "__ogre_constraints__");


        shareUpG->addNode(jean.id, jean);
        shareUpG->addNode(aime.id, aime);
        shareUpG->addNode(manger.id, manger);
        shareUpG->addNode(des.id, des);
        shareUpG->addNode(pommes.id, pommes);

        shareUpG->addEdge(aime.id, jean.id, suj);
        shareUpG->addEdge(aime.id, manger.id, obj);
        shareUpG->addEdge(manger.id, pommes.id, obj2);
        shareUpG->addEdge(pommes.id, des.id, det);
        shareUpG->addEdge(manger.id, aime.id, constraint);
    }

    void makeRedirectDownGraph()
    {
        redirectDownG = new LinguisticGraph;

        LinguisticNodeProperties node1(0, "Jean", "Jean", "N", "NPP"),
                node2(1, "aime", "aimer", "V", "V"),
                node3(2, "manger", "manger", "V", "VINF"),
                node4(3, "des", "le", "D", "DET"),
                node5(4, "pommes", "pomme", "N", "NC");

        LinguisticEdgeProperties edge1(0, "suj", "suj"),
                edge2(1, "obj", "obj"),
                edge3(2, "det", "det"),
                edge4(3, "obj", "obj");

        redirectDownG->addNode(node1.id, node1);
        redirectDownG->addNode(node2.id, node2);
        redirectDownG->addNode(node3.id, node3);
        redirectDownG->addNode(node4.id, node4);
        redirectDownG->addNode(node5.id, node5);

        // aime -suj-> Jean
        redirectDownG->addEdge(1, 0, edge1);
        // aime -obj-> manger
        redirectDownG->addEdge(1, 2, edge2);
        // manger -obj-> pommes
        redirectDownG->addEdge(2, 4, edge4);
        // pommes -det-> des
        redirectDownG->addEdge(4, 3, edge3);
    }

    void makeMoveDownGraph()
    {
        moveDownG = new LinguisticGraph;

        LinguisticNodeProperties jean(0, "Jean", "Jean", "N", "NPP"),
                aimer(1, "aime", "aimer", "V", "V"),
                marie(2, "Marie", "manger", "V", "VINF"),
                qui(3, "qui", "qui", "PRO", "PROREL");

        LinguisticEdgeProperties modrel(0, "mod.rel", "mod.rel"),
                //suj(1, "suj", "suj"),
                obj(2, "obj", "obj"),
                suj_real(3, "suj", "suj");

        moveDownG->addNode(jean.id, jean);
        moveDownG->addNode(aimer.id, aimer);
        moveDownG->addNode(marie.id, marie);
        moveDownG->addNode(qui.id, qui);

        // aime -suj-> qui
        //moveDownG->addEdge(aimer.id, qui.id, suj);
        // aime -obj-> marie
        moveDownG->addEdge(aimer.id, marie.id, obj);
        // jean -mod.rel-> aimer
        moveDownG->addEdge(jean.id, aimer.id, modrel);
        // aime -suj-> qui
        moveDownG->addEdge(aimer.id, qui.id, suj_real);
    }

    void makeAddFeatureGraph()
    {
        addFeatureG = new LinguisticGraph;
        LinguisticNodeProperties jean(0, "Jean", "Jean", "N", "NPP"),
                est (1, "est", "être", "V", "V"),
                elu(2, "élu", "élire", "V", "VPP");

        LinguisticEdgeProperties suj(0, "suj", ""),
                auxpass(1, "aux.pass", "aux.pass"),
                cons(2, "__constraints__", "__constraints__");


        addFeatureG->addNode(jean.id, jean);
        addFeatureG->addNode(est.id, est);
        addFeatureG->addNode(elu.id, elu);

        addFeatureG->addEdge(elu.id, jean.id, suj);
        addFeatureG->addEdge(elu.id, est.id, auxpass);
        addFeatureG->addEdge(elu.id, elu.id, cons);
    }

    LinguisticGraph *moveDownG,
                    *redirectDownG,
                    *shareUpG,
                    *shareDownG,
                    *addFeatureG;
};

typedef std::map<std::string, std::string> FeaturesMap;

BOOST_FIXTURE_TEST_SUITE(ConstraintSuite, ConstraintFixture)

BOOST_AUTO_TEST_CASE(ShareDownFinalTest){
    unsigned old_size = shareDownG->edgesSize();

    set<string> labels = boost::assign::list_of("obj");
    Constraint *shareDown = new ShareDown(labels, FeaturesMap() , DEBUG);

    //Constraint on edge number 2 (manger -obj-> pommes)
    //This:
    // aime -> manger -obj-> pommes
    //Will be rewritten as:
    // aime ( -> manger -obj-> pommes
    //        -obj> pommes)
    while(shareDown->trigger(shareDownG, 3))
        continue;


    BOOST_REQUIRE_EQUAL( shareDownG->edgesSize(), old_size+1);
    BOOST_CHECK_EQUAL( shareDownG->edgeProperties( 4 ).relFinal(), "obj" );

    BOOST_CHECK_EQUAL( shareDownG->nodeProperties(shareDownG->sourceNode(4)).lemma(), "aimer" );
    BOOST_CHECK_EQUAL( shareDownG->nodeProperties(shareDownG->targetNode(4)).lemma(), "pomme" );

}

BOOST_AUTO_TEST_CASE(ShareDownCanonicalTest){
    unsigned old_size = shareDownG->edgesSize();


    set<string> labels = boost::assign::list_of(":obj");
    Constraint *shareDown = new ShareDown(labels, FeaturesMap(), DEBUG);

    //Constraint on edge number 2 (manger -obj-> pommes)
    //This:
    // aime -> manger -obj-> pommes
    //Will be rewritten as:
    // aime ( -> manger -obj-> pommes
    //        -obj> pommes)
    while(shareDown->trigger(shareDownG, 3))
        continue;


    BOOST_REQUIRE_EQUAL( shareDownG->edgesSize(), old_size+1);
    BOOST_CHECK_EQUAL( shareDownG->edgeProperties( 4 ).relCanonical(), "obj" );

    BOOST_CHECK_EQUAL( shareDownG->nodeProperties(shareDownG->sourceNode(4)).lemma(), "aimer" );
    BOOST_CHECK_EQUAL( shareDownG->nodeProperties(shareDownG->targetNode(4)).lemma(), "pomme" );
}

BOOST_AUTO_TEST_CASE(ShareUpFinalTest){
    set<string> labels = boost::assign::list_of("suj");
    Constraint *shareUp = new ShareUp(labels, FeaturesMap(), DEBUG);

    //Constraint on edge number 4 (manger -[constraint]-> aime)
    //This :
    // manger -[cons]-> aime -[suj:suj]-> jean
    //Will be rewritten as
    // manger (
    //          -[cons]-> aime -[suj:suj]-> jean
    //          -[suj:suj]-> jean
    //        )
    while(shareUp->trigger(shareUpG, 4))
        continue;

    BOOST_CHECK_EQUAL( shareUpG->edgeProperties( 5 ).relFinal(), "suj" );

    bool found = false;
    LinguisticGraph::out_edge_iter out, out_end;
    for(boost::tie(out, out_end) = shareUpG->outEdges(2); out != out_end; ++out)
    {
        LinguisticEdgeProperties prop = shareUpG->edgeProperties(*out);
        if( shareUpG->nodeProperties(shareUpG->targetNode(*out)).id == 0 && prop.relFinal() == "suj" )
        {
            found = true;
            break;
        }
    }
    BOOST_CHECK_EQUAL( found, true );
}

BOOST_AUTO_TEST_CASE(ShareUpStatusTest){
    set<string> labels = boost::assign::list_of("suj");
    FeaturesMap features = boost::assign::map_list_of("surfacic", "n");
    Constraint *shareUp = new ShareUp(labels, features, DEBUG);

    //Constraint on edge number 4 (manger -[constraint]-> aime)
    //This :
    // manger -[cons]-> aime -[suj:suj]-> jean
    //Will be rewritten as
    // manger (
    //          -[cons]-> aime -[suj:suj]-> jean
    //          -[suj:suj]-> jean
    //        )
    while(shareUp->trigger(shareUpG, 4))
        continue;

    BOOST_CHECK_EQUAL( shareUpG->edgeProperties( 5 ).relFinal(), "suj" );
    BOOST_CHECK_EQUAL( shareUpG->edgeProperties( 5 ).type(), LinguisticEdgeProperties::Deep );
}

BOOST_AUTO_TEST_CASE(ShareUpCanonicalTest){    
    set<string> labels = boost::assign::list_of(":suj");
    Constraint *shareUp = new ShareUp(labels, FeaturesMap(), DEBUG);

    //Constraint on edge number 4 (manger -[constraint]-> aime)
    //This :
    // manger -[cons]-> aime -[suj:suj]-> jean
    //Will be rewritten as
    // manger (
    //          -[cons]-> aime -[suj:suj]-> jean
    //          -[suj:suj]-> jean
    //        )
    while(shareUp->trigger(shareUpG, 4))
        continue;

    BOOST_CHECK_EQUAL( shareUpG->edgeProperties( 5 ).relCanonical(), "suj" );

    bool found = false;
    LinguisticGraph::out_edge_iter out, out_end;
    for(boost::tie(out, out_end) = shareUpG->outEdges(2); out != out_end; ++out)
    {
        LinguisticEdgeProperties prop = shareUpG->edgeProperties(*out);
        if( shareUpG->nodeProperties(shareUpG->targetNode(*out)).id == 0 && prop.relCanonical() == "suj" )
        {
            found = true;
            break;
        }
    }
    BOOST_CHECK_EQUAL( found, true );
}

BOOST_AUTO_TEST_CASE(RedirectDownFinalTest){
    //Constraint on label "obj" only
    set<string> labels = boost::assign::list_of("obj");
    Constraint *redirectDown = new RedirectDown(labels, FeaturesMap(), DEBUG);

    //Constraint on edge number 3 (manger --> pommes)
    // This :
    // x -[obj]-> manger --> pommes
    // Will be rewritten as:
    // x -[obj]-> pommes <-- manger
    while(redirectDown->trigger(redirectDownG, 3))
        continue;

    BOOST_CHECK_EQUAL( redirectDownG->edgeProperties( 1 ).relFinal(), "obj" );

    bool found = false;
    LinguisticGraph::out_edge_iter out, out_end;
    for(boost::tie(out, out_end) = redirectDownG->outEdges(1); out != out_end; ++out)
    {
        LinguisticEdgeProperties prop = redirectDownG->edgeProperties(*out);
        if( redirectDownG->nodeProperties(redirectDownG->targetNode(*out)).id == 4 && prop.relFinal() == "obj" )
        {
            found = true;
            break;
        }
    }
    BOOST_CHECK_EQUAL( found, true );
}

BOOST_AUTO_TEST_CASE(RedirectDownCanonicalTest){
    //Constraint on label ":obj" only
    set<string> labels = boost::assign::list_of(":obj");
    Constraint *redirectDown = new RedirectDown(labels, FeaturesMap(), DEBUG);

    //Constraint on edge number 3 (label obj between manger and pommes)
    while(redirectDown->trigger(redirectDownG, 3))
        continue;

    BOOST_CHECK_EQUAL( redirectDownG->edgeProperties( 1 ).relCanonical(), "obj" );

    bool found = false;
    LinguisticGraph::out_edge_iter out, out_end;
    for(boost::tie(out, out_end) = redirectDownG->outEdges(1); out != out_end; ++out)
    {
        LinguisticEdgeProperties prop = redirectDownG->edgeProperties(*out);
        if( redirectDownG->nodeProperties(redirectDownG->targetNode(*out)).id == 4 && prop.relCanonical() == "obj" )
        {
            found = true;
            break;
        }
    }
    BOOST_CHECK_EQUAL( found, true );
}

BOOST_AUTO_TEST_CASE(MoveDownFinalTest){
    //DFA to find mod.rel
    unordered_set<string> alphabet = boost::assign::list_of("suj")("obj")("mod.rel");
    DFA *dfa = new DFA(alphabet);
    dfa->addTransition(0, "__not__mod.rel", 0);
    dfa->addTransition(0, "mod.rel", 1);
    dfa->addFinalState(1);

    Constraint *moveDown = new MoveDown(dfa, "", DEBUG);

    //Constraint on edge number 3 (aimer -suj-> qui)
    while(moveDown->trigger(moveDownG, 3))
        continue;

    bool found = false;
    LinguisticGraph::out_edge_iter out, out_end;
    for(boost::tie(out, out_end) = moveDownG->outEdges(1); out != out_end; ++out)
    {
        LinguisticEdgeProperties prop = moveDownG->edgeProperties(*out);
        if( moveDownG->nodeProperties(moveDownG->targetNode(*out)).id == 0 && prop.relFinal() == "suj" )
        {
            found = true;
            break;
        }
    }
    BOOST_CHECK_EQUAL( found, true );
}

BOOST_AUTO_TEST_CASE(AddFeatureTest){
    set<string> labels = boost::assign::list_of("aux.pass");
    Constraint *addFeature = new AddFeature(labels, "suj", "obj", FeaturesMap(), DEBUG);

    //Constraint on constraint edge cons (elu -__constraints__-> elu)
    while(addFeature->trigger(addFeatureG, 2))
        continue;

    BOOST_CHECK_EQUAL(addFeatureG->edgeProperties(0).relFinal(), "suj");
    BOOST_CHECK_EQUAL(addFeatureG->edgeProperties(0).relCanonical(), "obj");
}

BOOST_AUTO_TEST_SUITE_END()
