#include <boost/test/unit_test.hpp>
#include <boost/assign/list_of.hpp>
#include <iostream>

#include "ogrecomponents/dfa.h"

using namespace std;

using namespace ogrecomponents;

struct DFAFixture
{
public:
    DFAFixture()
    {
        dfa = new DFA(unordered_set<string>());
    }

    DFAFixture(unordered_set<string> alpha)
    {
        dfa = new DFA(alpha);
    }

    void changeAlphabet(unordered_set<string> alphabet)
    {
        delete dfa;
        dfa = NULL;
        dfa = new DFA(alphabet);
    }

    DFA *dfa;
};


BOOST_FIXTURE_TEST_SUITE(DFASuite, DFAFixture)

BOOST_AUTO_TEST_CASE(DFACreationTest){
    dfa->addTransition(0, "test", 1);
    dfa->addFinalState(1);

    BOOST_CHECK_EQUAL(dfa->isFinal(1), true);
    BOOST_CHECK_EQUAL(dfa->testTransition("test"), 1);
}

BOOST_AUTO_TEST_CASE(DFANotTest)
{
    unordered_set<string> alphabet = boost::assign::list_of("test")("test2")("test3");
    changeAlphabet(alphabet);

    dfa->addTransition(0, "__not__test", 1);
    dfa->addFinalState(1);

    BOOST_CHECK_EQUAL(dfa->isFinal(1), true);
    BOOST_CHECK_EQUAL(dfa->testTransition("test2"), 1);
}

BOOST_AUTO_TEST_CASE(DFATestDisjunction)
{
    dfa->addTransition(0, "test|test2|test3", 1);
    dfa->addFinalState(1);

    BOOST_CHECK_EQUAL(dfa->testTransition("test"), 1);
    BOOST_CHECK_EQUAL(dfa->testTransition("test2"), 1);
    BOOST_CHECK_EQUAL(dfa->testTransition("test3"), 1);
}

BOOST_AUTO_TEST_CASE(DFARecognitionTest)
{
    dfa->addTransition(0, "test2", 1);
    dfa->addTransition(0, "test", 2);
    dfa->addTransition(2, "ok", 3);
    dfa->addFinalState(3);

    BOOST_CHECK_EQUAL(dfa->isFinal(3), true);

    vector<unsigned> path, expected_path = boost::assign::list_of(0)(2)(3);
    vector<string> input = boost::assign::list_of("ok")("test"); //use like a stack

    BOOST_CHECK_EQUAL(dfa->start(), 0);

    path.push_back(dfa->start());
    while(!dfa->isFinal(dfa->current()) && input.size() > 0)
    {
        string label = input.back();
        input.pop_back();
        try{
            dfa->testTransition(label);
            path.push_back(dfa->next(label));
        }catch(const DFAError &e)
        {
            break;
        }
    }

    BOOST_CHECK_EQUAL_COLLECTIONS(path.begin(), path.end(),
                                  expected_path.begin(), expected_path.end());
}

BOOST_AUTO_TEST_SUITE_END()
